import pathlib as p
import sys
import threading as th
import traceback
import warnings as ws
from collections.abc import Callable

from colored import Back, Fore, Style

from .tcr_classfuncs import get_classname
from .tcr_console import console
from .tcr_iterable import cut_at
from .tcr_print import FMTC

C_WHITE = Fore.white + Style.bold
C_NUMBER = FMTC.NUMBER
C_WARNING = Fore.yellow + Style.bold
C_ERROR = Fore.red + Style.bold
C_EXCEPTION = FMTC.INTERNAL_EXCEPTION


def default_err_or_warning_printhook(s: str, warning: Warning | Exception, filename: str, lineno: int, *, tb=None) -> None:
  """Default printhook for tcr.WarningCatcher."""
  path = p.Path(filename)

  AROUND: int = 2
  """How many lines around the line of the warning to show."""
  CUTOFF: int = 70
  """How many (max) characters can be shown per line until they ge..."""

  is_warning = isinstance(warning, Warning)

  color = C_WARNING if is_warning else C_ERROR

  if not path.exists():
    lines = ''
  else:
    lines = path.read_text().split('\n')
    lines = lines[max(0, lineno - (AROUND + 1)) : lineno + AROUND]
    while all(x.startswith(' ') for x in lines):
      lines = [x[1:] for x in lines]
    lines = [f'{C_NUMBER}{lineno-AROUND+i}{color} {">" if i == AROUND else "|"}{FMTC._} {cut_at(x, CUTOFF)}' for i, x in enumerate(lines)]
    lines = '\n'.join(lines)
    lines = f'\n{lines}'

  if tb:
    traceback.print_tb(tb)

  func = console.warn if is_warning else console.error
  func(f'{color}{filename}{C_WHITE}:{C_NUMBER}{lineno}{C_WHITE}: {C_EXCEPTION}{get_classname(warning)}{C_WHITE}: {color}{str(warning).replace("<locals>.", "")}{lines}')


def _errorcatcher_excepthook(type, value, traceback):
  frame = traceback.tb_frame
  filename = frame.f_code.co_filename
  lineno = frame.f_lineno
  default_err_or_warning_printhook(str(value), value, filename, lineno, tb=traceback)


class ErrorCatcher:
  def __init__(self, **kwargs) -> None:
    self.install(**kwargs)

  @staticmethod
  def install():
    sys.excepthook = _errorcatcher_excepthook


class WarningCatcher:
  """Catches any warnings (for example "Coroutine X was not awaited") and forwards them as text to the selected (or default: tcr.console.warn) printhook.

  Usage:
  ```py
  tcr.WarningCatcher()
  # OR
  tcr.WarningCatcher.install() # Tgus
  ```
  """

  @staticmethod
  def install(**kwargs):
    """Alias for tcr.WarningCatcher(**kwargs). Use it for better typehints."""
    _WarningCatcher.install(**kwargs)

  def __init__(
    self,
    printhook: Callable[[str, Warning, str, int], None] = default_err_or_warning_printhook,
  ) -> None:
    """Catches any warnings (for example "Coroutine X was not awaited") and forwards them as text to the selected (or default: tcr.console.warn) printhook.

    Usage:
    ```py
    tcr.WarningCatcher()
    ```

    Args:
        - printhook: Callable[[str, Warning, str, int], None], Defines a printer function, for more info see this module's default_showwarning_printhook function and https://docs.python.org/3/library/warnings.html#warnings.showwarning.
    """
    _WarningCatcher.install(
      printhook=printhook,
    )


class _WarningCatcher(th.Thread):
  _initiated = False
  _installed = False

  def __init__(self, *, printhook):
    if _WarningCatcher._initiated:
      raise RuntimeError('WarningCatcher can only be initiated once.')
    super().__init__()
    _WarningCatcher._initiated = True
    self.__printhook = printhook
    self.daemon = True
    self._stop_event = th.Event()

  def run(self):
    def showwarning(message, category, filename, lineno, file=None, line=None):
      self.__printhook(str(message), message, filename, lineno)

    with ws.catch_warnings():
      ws.showwarning = showwarning
      while not self._stop_event.is_set():
        self._stop_event.wait(0.1)  # Check every 0.1 second for the stop signal

  def stop(self):
    self._stop_event.set()

  @classmethod
  def install(cls, **kwargs):
    if _WarningCatcher._installed:
      raise RuntimeError('WarningCatcher can only be installed once.')
    _WarningCatcher._installed = True

    catcher = cls(**kwargs)
    catcher.start()
    return catcher



from __future__ import annotations as __future_annotations

from ...discord.tcrd_snowflake import is_snowflake as __is_snowflake
from ...discord.tcrd_string import IFYs
from ...src.tcr_console import console as c
from .. import util as __

if True:  # Author

  @__.REQUIRE('ctx', 'event')
  @__.REQUIRE('user_mentions')
  @__.SWITCH('noping')
  def mention(
    name,
    user: str | None = None,
    *args: str,
    ctx: __.arc.Context | None,
    event: __.hikari.MessageCreateEvent | None,
    user_mentions: list,
    noping: bool,
    **ctxs,
  ) -> str:
    """### Return user's mention.

    ### Example:
    ```txt
    Hello {@}!
    ```

    ### Result:
    ```txt
    Hello <@507642999992352779>!
    ```

    ### Dependencies
    (SlashContext || MessageCreateEvent) && user_mentions
    """

    if user is None:
      author_id = ctx.author.id if ctx else event.author.id
    else:
      if __is_snowflake(user):
        author_id = int(user)
      else:
        return __.rebuild_yourself(name, user, *args, **ctxs)

    if not noping and author_id not in user_mentions:
      user_mentions.append(author_id)

    return IFYs.userify(author_id)

  @__.REQUIRE_AUTHOR
  def username(
    *args,
    author: __.hikari.User,
    **ctxs,
  ) -> str:
    """### Return user's username.

    Example:
    ```txt
    Hello {username}!
    ```

    Result:
    ```txt
    Hello anamoyee!
    ```

    ### Dependencies
    (SlashContext || MessageCreateEvent) && user_mentions
    """
    return author.username

  @__.REQUIRE_AUTHOR
  def globalname(
    *args,
    author: __.hikari.User,
    **ctxs,
  ):
    """### Return user's global name.

    Example:
    ```txt
    Hello {globalname}
    ```

    Result:
    ```txt
    Hello anamoyee :3
    ```

    ### Dependencies
    SlashContext || MessageCreateEvent
    """
    return author.global_name

  @__.REQUIRE_MEMBER_AUTHOR(lambda *_, author, **__: author.global_name or author.username)
  def nickname(
    *args,
    author: __.hikari.User,
    member: __.hikari.Member,
    **ctxs,
  ) -> str:
    """### Return user's nickname, then global name, then username if any are not available.

    Return various user's names in that order
     - `nickname` if they have one, otherwise:
     - `globalname` if they have one, otherwise:
     - `username`

    Example:
    ```txt
    Hello {nickname}
    ```

    Result:
    ```txt
    Hello anamoyee [🦊🎴]
    ```

    ### Dependencies
    (SlashContext && InGuild) || GuildMessageCreateEvent
    """
    return member.nickname or author.global_name or author.username

  @__.STRINGIFY
  @__.REQUIRE_AUTHOR
  def discriminator(
    *args,
    author: __.hikari.User,
    **ctxs,
  ) -> str:
    """### Return user's discriminator.

    Example:
    ```txt
    Hello {discriminator}!
    ```

    Result:
    ```txt
    Hello 0009!
    ```

    ### Dependencies
    SlashContext || MessageCreateEvent
    """

    if hasattr(author, 'discrim'):
      return author.discriminator
    else:
      return 0

  @__.REQUIRE_AUTHOR
  def tag(
    *args,
    author: __.hikari.User,
    **ctxs,
  ) -> str:
    """### Return user's tag.

    Example:
    ```txt
    Hello {tag}!
    ```

    Result:
    ```txt
    Hello anamoyee#0009! // Whenever user's discriminator != 0
    ```
    OR
    ```txt
    Hello @anamoyee // Whenever user's discriminator == 0
    ```

    ### Dependencies
    SlashContext || MessageCreateEvent
    """
    discrim = author.discriminator if hasattr(author, 'discrim') else '0'

    if discrim == '0':
      return f'@{author.username}'
    else:
      return f'{author.username}#{discrim}'

  @__.REQUIRE_AUTHOR
  def id(
    *args,
    author: __.hikari.User,
    **ctxs,
  ) -> str:
    """### Return user's id.

    Example:
    ```txt
    Hello {id}!
    ```

    Result:
    ```txt
    Hello 507642999992352779!
    ```

    ### Dependencies
    SlashContext || MessageCreateEvent
    """
    return str(author.id)

  @__.REQUIRE_AUTHOR
  def bot(
    *args,
    author: __.hikari.User,
    **ctxs,
  ):
    """### Return whether user is bot or not.

    Example:
    ```txt
    It is {bot} that you are a bot!
    ```

    Result:
    ```txt
    It is false that you are a bot!
    ```

    ### Dependencies
    SlashContext || MessageCreateEvent
    """
    return __.jsbool(author.is_bot)

  @__.REQUIRE('event', 'ctx')
  def human(
    *args,
    ctx: __.arc.Context | None = None,
    event: __.hikari.MessageCreateEvent | None = None,
    **ctxs,
  ):
    """### Return whether user is the system user or not.

    Example:
    ```txt
    It is {system} that you are the system user!
    ```

    Result:
    ```txt
    It is false that you are the system user!
    ```

    ### Dependencies
    SlashContext || MessageCreateEvent
    """
    if ctx:
      return 'true'
    else:
      return __.jsbool(event.is_human)

  @__.REQUIRE('event', 'ctx')
  def in_dms(
    *args,
    ctx: __.arc.Context | None = None,
    event: __.hikari.MessageCreateEvent | None = None,
    **ctxs,
  ):
    """### Return whether user is in DMs or not.

    Example:
    ```txt
    It is {indms} that you are running this command in DMs!
    ```

    Result:
    ```txt
    It is true that you are running this command in DMs!
    ```

    ### Dependencies
    SlashContext || MessageCreateEvent
    """
    return __.jsbool(not hasattr(ctx or event, 'guild_id'))

  @__.REQUIRE_AUTHOR
  def avatar(
    *args,
    author: __.hikari.User,
    **ctxs,
  ) -> str:
    """### Return user's avatar url.

    Example:
    ```txt
    Hello {avatar}
    ```

    Result:
    ```txt
    Here's your avatar url https://cdn.discordapp.com/avatars/507642999992352779/cef9fec7c8afce0635c24dff93ff6010.png?size=4096
    ```

    ### Dependencies
    SlashContext || MessageCreateEvent
    """
    return str(author.display_avatar_url)

  @__.REQUIRE_MEMBER_AUTHOR('0')
  def roles(
    *args,
    member: __.hikari.Member,
    **ctxs,
  ) -> str:
    """### Return user's roles.

    Example:
    ```txt
    You have {roles} roles!
    ```

    Result:
    ```txt
    You have 6 roles!
    ```

    ### Dependencies
    (SlashContext && InGuild) || GuildMessageCreateEvent
    """
    return str(len(member.role_ids) - 1)  # -1 to compensate for the @everyone role

  @__.REQUIRE('attachments')
  @__.REQUIRE_POSITIONAL(1, None)
  def attach(
    _,
    url: str,
    *args,
    attachments: list[__.hikari.Attachment],
    **ctxs,
  ):
    """### Attach a file to the message from a given url.

    This will wieve out the url if it's not a valid url,
    however, this may cause a BadRequestError if the url is invalid.

    Example:
    ```txt
    {attach|{avatar}}
    ```

    Result:
    ```txt
    <no content but the avatar will be attached as file>
    ```

    ### Dependencies
    attachments
    """
    url = url.strip()
    if not __.regex.match(__.RegexPreset.URL, url):
      return f"Something went wrong while attaching a file, here's a link instead: <{url}>"

    attachments.append(url)

  @__.SWITCH('nohash')
  @__.REQUIRE_MEMBER_AUTHOR(lambda *_, nohash, **__: f'{"" if nohash else "#"}000000')
  def color(
    *args,
    member: __.hikari.Member,
    nohash: bool = False,
    **ctxs,
  ) -> str:
    """### Return user's topmost role's color (skipping non-colored roles).

    Example:
    ```txt
    Your color is {color}
    ```

    Result:
    ```txt
    Your color is #ff8000
    ```

    ### Dependencies
    (SlashContext && InGuild) || GuildMessageCreateEvent
    """
    top_color = 0x0
    for role in member.get_roles():
      if str(role.color) != '#000000':
        top_color = int(role.color)
        break

    return f'{__.tcrhex(top_color, leading_zeroes=6, prefix=("" if nohash else "#"))}'


if True:  # Server

  async def server(
    name: str,
    subname: str,
    *args: str,
    **ctxs,
  ) -> str:
    """### Dispatcher for all server-related subplaceholders.

    Example:
    ```txt
    Server name: {server|name}{#|You have to specify the subplaceholder 'name'}
    ```

    Result:
    ```txt
    Server name: Mt. Celeste Paceping Association
    ```

    ### Dependencies
    Technically None but all subplaceholders require (SlashContext && InGuild) || GuildMessageCreateEvent
    """
    LOOKUP = {
      'name': _server_name,
      'id': _server_id,
    }

    if subname in LOOKUP:
      return await __.run_sac(LOOKUP[subname], name, subname, *args, **ctxs)
    else:
      return __.rebuild_yourself(name, subname, *args, **ctxs)

  @__.REQUIRE_GUILD()
  def _server_name(
    *args,
    guild: __.hikari.Guild,
    **ctxs,
  ):
    """### Get the name of the server.

    Example:
    ```txt
    Server name: {server|name}{#|You have to specify the subplaceholder 'name'}
    ```

    Result:
    ```txt
    Server name: Mt. Celeste Paceping Association
    ```

    ### Dependencies
    (SlashContext && InGuild) || GuildMessageCreateEvent
    """
    return guild.name

  @__.REQUIRE_GUILD()
  def _server_id(
    *args,
    guild: __.hikari.Guild,
    **ctxs,
  ):
    return str(guild.id)



"""Functions that are supposed to be run in development for example to generate a type hinter."""

import inspect
from types import FunctionType
from typing import Any

from .tcr_classfuncs import get_name_classname

print_ = print


def generate_function_argument_typehints(func: FunctionType) -> str:
  try:
    signature = inspect.signature(func)
  except ValueError:  # Signature not found
    return ''
  arguments_with_typehints = []
  for param in signature.parameters.values():
    if param.kind == inspect.Parameter.VAR_POSITIONAL:
      arguments_with_typehints.append('*' + param.name)
    elif param.kind == inspect.Parameter.VAR_KEYWORD:
      arguments_with_typehints.append('**' + param.name)
    else:
      argument = param.name
      if param.annotation != inspect.Parameter.empty:
        argument += f': {get_name_classname(param.annotation)}'
      if param.default == inspect.Parameter.empty:
        arguments_with_typehints.append(argument)
      else:
        arguments_with_typehints.append(f'{argument} = {param.default!r}')
  return ', '.join(arguments_with_typehints)


def generate_function_return_typehint(func: FunctionType) -> str:
  try:
    signature = inspect.signature(func)
  except ValueError:  # Signature not found
    return ''
  return_type = signature.return_annotation
  if return_type == inspect.Signature.empty:
    return ''
  else:
    return f' -> {get_name_classname(return_type)}'


def generate_type_hinter(
  obj: Any,
  /,
  *,
  clipboard: bool = False,
  print: bool = False,
  skip_leading_underscore: bool = False,
  attrib_whitelist: tuple[str] = (),
  attrib_blacklist: tuple[str] = ('sys', 'os', 'warnings'),
) -> str | None:
  """Generates a string class definition for typehinting based on the object's dir() and the types of its attributes.

  Args:
    obj (Any): The object to generate the type hint for.
    clipboard (bool, optional): Whether to copy the result to the clipboard. Defaults to False. (Will also suppress any output to not flood the console)
    print (bool, optional): Whether to print the result. Defaults to True. (Will also suppress any output to not flood the console)
    skip_leading_underscore (bool, optional): Whether to skip leading underscores in the generated type hints. When False, skip only __double_leading_underscore (except some).
    attrib_whitelist (tuple[str], optional): A tuple of attribute names to include in the type hint even if they don't match the specified `skip_leading_underscore` value (or none at all). Defaults to ().
    attrib_blacklist (tuple[str], optional): A tuple of attribute names to exclude from the type hint no matter what. Defaults to ('sys', 'os').
  """

  def sort_key(x):
    v = getattr(obj, x)
    if inspect.isclass(v):
      return 3
    if inspect.ismodule(v):
      return 2
    elif callable(v):
      return 1
    else:
      return 0

  hints = []
  processed_hints_dict = {}
  valid = (lambda x: not x.startswith('_')) if skip_leading_underscore else lambda x: not x.startswith('__')
  obj_dir = sorted([x for x in dir(obj) if ((valid(x)) or x in attrib_whitelist) and x not in attrib_blacklist], key=sort_key)
  for attr_name in obj_dir:
    attr_value = getattr(obj, attr_name)
    attr_type = get_name_classname(attr_value)
    if inspect.isclass(attr_value) or inspect.ismodule(attr_value):
      hints.append('\n  '.join(generate_type_hinter(attr_value).split('\n')))
    elif callable(attr_value):
      args_str = generate_function_argument_typehints(attr_value)
      hints.append(f'@staticmethod\n  def {attr_name}({args_str}){generate_function_return_typehint(attr_value)}: ...')
    else:
      hints.append(f'{attr_name}: {attr_type}')
    processed_hints_dict[attr_name] = attr_value

  if not obj_dir:
    hints.append('...')

  # hint types (imports)

  result = f'class {get_name_classname(obj).split(".")[-1]}:\n' + '\n'.join(f'  {x}' for x in hints)

  if print:
    if callable(print):
      print(result)
    else:
      print_(result)

  if clipboard:
    from .tcr_console import console

    try:
      import pyperclip
    except ImportError:
      print('\npyperclip not installed. Installing...')
      import subprocess as __sp

      try:
        __sp.check_call(['pip', 'install', 'pyperclip'])
      except __sp.CalledProcessError:
        console.error('Unable to install pyperclip... Install it manually, returning without copy')
      else:
        import pyperclip

        pyperclip.copy(result)
        console.log(f'Copied to clipboard! ({len(result)} characters)')
      del __sp
    else:
      pyperclip.copy(result)
      console.log(f'Copied to clipboard! ({len(result)} characters)')

  if not (clipboard or print):
    return result
  else:
    return None



import datetime


class UptimeResult:
  string: str
  """Formatted string, for example: `10:20:30`"""
  total_seconds: int
  """Total seconds, for example: 19234, For parsed seconds use `s`"""
  d: int
  """Days"""
  h: int
  """Hours"""
  m: int
  """Minutes"""
  s: int
  """Seconds, for total seconds use `total_seconds`"""
  delta: datetime.timedelta

  def __init__(
    self,
    *,
    string: str,
    total_seconds: int,
    d: int,
    h: int,
    m: int,
    s: int,
    delta: datetime.timedelta,
  ) -> None:
    self.string = string
    self.total_seconds = total_seconds
    self.d = d
    self.h = h
    self.m = m
    self.s = s
    self.delta = delta

  def __int__(self) -> int:
    return self.total_seconds

  def __str__(self) -> str:
    return self.string


class Uptime:
  """Counts time from when it's initialised. Returns as formatted str, datetime.delta, total_seconds or each number component (days, hours, minutes, seconds)."""

  def __init__(self) -> None:
    self.__start_time = datetime.datetime.now(tz=datetime.UTC)

  def get(self) -> UptimeResult:
    delta = datetime.datetime.now(tz=datetime.UTC) - self.__start_time
    total_seconds = delta.total_seconds()
    d = total_seconds // 86400
    h = (total_seconds // 3600) % 24
    m = (total_seconds // 60) % 60
    s = total_seconds % 60
    if d > 0:
      fmted = f'{int(d)}d + {int(h):02d}:{int(m):02d}:{int(s):02d}'
    else:
      fmted = f'{int(h):02d}:{int(m):02d}:{int(s):02d}'
    return UptimeResult(
      string=fmted,
      total_seconds=int(total_seconds),
      d=d,
      h=h,
      m=m,
      s=s,
      delta=delta,
    )

  def __str__(self) -> str:
    return str(self.get())

  def __int__(self) -> int:
    return int(self.get())



"""Functions related to comparisons, didn't want to name the file tcr_test.py to avoid confusion with testing as in testing if code works."""

from collections.abc import Callable
from typing import Any


class AbleResult:
  """Contains a result from tcr.able().

  May be used in an if statement or bool(able_result) expression, in whichc ase
  """

  result: Any | BaseException
  truefalse: bool
  """Can't name it bool, can't name it result, can't name it anything... bruh. truefalse it is then"""

  def __init__(self, truefalse: bool, result: Any):
    self.truefalse = truefalse
    self.result = result

  def __bool__(self) -> bool:
    return self.truefalse

  def __iter__(self):
    yield self.truefalse
    yield self.result


def able(func: Callable, *args, able_exception__: BaseException = Exception, **kwargs) -> AbleResult:
  """### Return True* if x is func(x)-able aka does not raise any errors while passed in to func otherwise return False*.

  *The result type is AbleResult, not bool, but it's bool-able (can be used in an if statement or bool(able_result) normally)

  #### Example:
  ```py
  >>> able(int, '1') # Wouldn't raise any Exception
  True
  >>> able(int, 'a') # Would raise: ValueError: invalid literal for int() with base 10: 'a'
  False
  >>> [int(x) for x in [] if able(int, x)]
  ```

  #### Result retrieving example:
  ```py
  >>> color = 'ff8000' # a hex encoded color code
  >>> if (r := able(int, color, base=16)):
  ...   print(repr(r.result))
  16744448
  >>> color = 'ff8000'
  >>> if (r := able(int, color)) or True: # oops! forgot to set the base (and added or True so this example actually prints)
  ...   print(repr(r.result)) # Returns False along with the exception in the result field
  ValueError("invalid literal for int() with base 10: 'ff8000'")
  ```
  """
  try:
    result = func(*args, **kwargs)
  except able_exception__ as e:
    return AbleResult(False, e)
  else:
    return AbleResult(True, result)


def isdunder(s: str):
  """Return True if the string ends and starts with at least two underscores each.

  A 2-char string containing just underscores is considered dunder.
  """

  return s.startswith('__') and (s.endswith('__'))



import pathlib as p
from collections.abc import Callable
from sys import exit
from typing import ParamSpec, Self

try:
  import glfw as gf
  import imgui
  import OpenGL.GL as gl
  from imgui.integrations.glfw import GlfwRenderer
except ModuleNotFoundError:
  pass
else:

  def ctrlc_grace_exit() -> bool:
    print('^C', end='')
    return True  # Exit

  P = ParamSpec('P')

  class ImGuiHandler:
    frame: Callable[P, None]
    """The frame callback."""
    title: str
    """The title of the window."""
    font_path: p.Path | None
    """The path to the font file to use, or None to disable."""
    ctrlc_handler: Callable[[], bool]
    """The Ctrl-C handler to use. Return True to exit the gui, False to continue. Do not exit()."""

    def __init__(self):
      self.title = 'Untitled window'
      self.font_path = None
      self.ctrlc_handler = ctrlc_grace_exit
      self.start_maximised = False

    def set_title(self, title: str = 'Untitled window') -> Self:
      if not isinstance(title, str):
        raise TypeError('title must be str')
      self.title = title
      return self

    def set_font_path(self, font_path: p.Path | str | None = None) -> Self:
      if isinstance(font_path, str):
        font_path = p.Path(font_path)
      if font_path is not None and not isinstance(font_path, p.Path):
        raise TypeError('font_path must be str or pathlib.Path (or None to disable)')
      if font_path is not None and not font_path.is_file():
        raise FileNotFoundError(f"font_path {font_path} doesn't exist or is not a file")
      self.font_path = font_path
      return self

    def set_ctrlc_handler(self, ctrlc_handler: Callable[[], bool] = ctrlc_grace_exit) -> Self:
      if ctrlc_handler is not None and not callable(ctrlc_handler):
        raise TypeError('ctrlc_handler must be Callable[[], bool]')
      self.ctrlc_handler = ctrlc_handler
      return self

    def set_start_maximised(self, start_maximised: bool = False) -> Self:
      self.start_maximised = start_maximised
      return self

    def __call__(self, frame: Callable[P, None], /):
      self.frame = frame
      return self

    def run(self, *args: P.args, **kwargs: P.kwargs) -> None:
      self.args = args
      self.kwargs = kwargs
      imgui.create_context()
      if not gf.init():
        raise RuntimeError('Could not initialize OpenGL context')

      gf.window_hint(gf.CONTEXT_VERSION_MAJOR, 3)
      gf.window_hint(gf.CONTEXT_VERSION_MINOR, 3)
      gf.window_hint(gf.OPENGL_PROFILE, gf.OPENGL_CORE_PROFILE)
      gf.window_hint(gf.OPENGL_FORWARD_COMPAT, gl.GL_TRUE)

      window = gf.create_window(*gf.get_monitor_physical_size(gf.get_primary_monitor()), self.title, None, None)
      gf.make_context_current(window)

      if not window:
        gf.terminate()
        raise RuntimeError('Could not initialize GLFW window.')

      if self.start_maximised:
        gf.maximize_window(window)

      impl = GlfwRenderer(window)

      io = imgui.get_io()
      jb = io.fonts.add_font_from_file_ttf(self.font_path, 30) if self.font_path is not None else None
      impl.refresh_font_texture()

      while not gf.window_should_close(window):
        try:
          self._render(impl, window, jb)
        except KeyboardInterrupt:
          if self.ctrlc_handler():
            break

      impl.shutdown()
      gf.terminate()

    def _render(self, impl, window, font):
      gf.poll_events()
      impl.process_inputs()
      imgui.new_frame()

      gl.glClearColor(0.1, 0.1, 0.1, 1)
      gl.glClear(gl.GL_COLOR_BUFFER_BIT)

      if font is not None:
        imgui.push_font(font)

      self.frame(*self.args, **self.kwargs)

      if font is not None:
        imgui.pop_font()

      imgui.render()
      impl.render(imgui.get_draw_data())
      gf.swap_buffers(window)



import inspect
import time
from asyncio import iscoroutinefunction
from collections.abc import Callable
from functools import partial, wraps

from colored import Back, Fore, Style

from .tcr_print import print_block

if True:  # \/ # @test

  def test(func: Callable) -> Callable:
    """A decorator for adding test label."""

    @wraps(func)
    def wrapper(*args, **kwargs):
      name = func.__name__
      name = name.removeprefix('test_')
      name = name.replace('_', ' ')
      name = name.title()
      print_block(name)
      return func(*args, **kwargs)

    return wrapper


if True:  # \/ # @timeit // timeit.start() and .stop()

  class Timeit:
    t: float | None
    tname: str

    def __init__(self) -> None:
      self.t = {}

    def _print_time(self, printhook, pattern, name, elapsed_time, color):
      printhook(
        pattern
        % {
          'name': name,
          'time': f'{elapsed_time:.4f}',
          'c_White': (Fore.white + Style.bold) if color else '',
          'c_Gold': (Fore.yellow + Style.bold) if color else '',
          'c_reset': Style.reset if color else '',
        }
      )

    def start(self, name='test'):
      self.t[name] = time.perf_counter()

    def stop(
      self,
      name='test',
      *,
      printhook=print,
      color=True,
      pattern='%(c_White)s%(name)s%(c_Gold)s took %(c_White)s%(time)s%(c_Gold)s seconds to execute.%(c_reset)s',
    ):
      if self.t.get(name) is None:
        msg = 'You cannot stop the timer if it was not started'
        raise RuntimeError(msg)
      diff = time.perf_counter() - self.t[name]
      self._print_time(printhook, pattern, name, diff, color)

    def __call__(
      self,
      func=None,
      *,
      printhook=None,
      color=True,
      pattern='%(c_White)s%(name)s%(c_Gold)s took %(c_White)s%(time)s%(c_Gold)s seconds to execute.%(c_reset)s',
    ):
      if func is None:
        return partial(timeit, printhook=printhook, pattern=pattern, color=color)
      if printhook is None:
        printhook = print

      @wraps(func)
      def wrapper(*args, **kwargs):
        start_time = time.perf_counter()
        result = func(*args, **kwargs)
        end_time = time.perf_counter()
        elapsed_time = end_time - start_time
        self._print_time(printhook, pattern, func.__name__, elapsed_time, color)
        return result

      return wrapper

  timeit = Timeit()

if True:  # \/ # @autorun, @instance

  def autorun(func):
    func()
    return func

  def instance(func):
    return func()


if True:  # \/ # @convert.stringify

  class Convert:
    def stringify(self, func):
      @wraps(func)
      def wrapper(*args, **kwargs):
        return str(func(*args, **kwargs))

      return wrapper

    def intify(self, func):
      @wraps(func)
      def wrapper(*args, **kwargs):
        return int(func(*args, **kwargs))

      return wrapper

    def floatify(self, func):
      @wraps(func)
      def wrapper(*args, **kwargs):
        return float(func(*args, **kwargs))

      return wrapper

    def boolify(self, func):
      @wraps(func)
      def wrapper(*args, **kwargs):
        return bool(func(*args, **kwargs))

      return wrapper

    def listify(self, func):
      @wraps(func)
      def wrapper(*args, **kwargs):
        return list(func(*args, **kwargs))

      return wrapper

    def setify(self, func):
      @wraps(func)
      def wrapper(*args, **kwargs):
        return set(func(*args, **kwargs))

      return wrapper

    def dictify(self, func):
      @wraps(func)
      def wrapper(*args, **kwargs):
        return dict(func(*args, **kwargs))

      return wrapper

    def tuplify(self, func):
      @wraps(func)
      def wrapper(*args, **kwargs):
        return tuple(func(*args, **kwargs))

      return wrapper

    def complexify(self, func):
      @wraps(func)
      def wrapper(*args, **kwargs):
        return complex(func(*args, **kwargs))

      return wrapper

    def reprify(self, func):
      @wraps(func)
      def wrapper(*args, **kwargs):
        return repr(func(*args, **kwargs))

      return wrapper

    def anyify(self, converter: Callable):
      def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
          return converter(func(*args, **kwargs))

        return wrapper

      return decorator

  convert = Convert()

  def aconvert(converter: Callable, await_converter: bool = False):
    def decorator(afunc):
      @wraps(afunc)
      async def wrapper(*args, **kwargs):
        if await_converter:
          return await converter(await afunc(*args, **kwargs))
        else:
          return converter(await afunc(*args, **kwargs))

      return wrapper

    return decorator


if True:  # \/ # @printer

  def printer(*, printhook: Callable[[str], None] = print, passthrough: bool = True):
    def decorator(func: Callable):
      @wraps(func)
      def wrapper(*args, **kwargs):
        result = func(*args, **kwargs)
        printhook(result)
        if passthrough:
          return result

      return wrapper

    return decorator


if True:  # \/ # @copy_kwargs

  def copy_kwargs(func):
    """Pack all of that func's kwargs into __kwargs and pass the kwarguments normally along with __kwargs kwarg which contains all the kwargs in a dict.

    This will error if func receives a literal "__kwargs" kwarg.
    ```py
    >>> a(__kwargs=kwargs, **{"__kwargs": "some_value"})
    ```
    ```txt
    TypeError: __main__.a() got multiple values for keyword argument '__kwargs'
    ```
    """

    @wraps(func)
    def wrapper(*args, **kwargs):
      return func(*args, __kwargs=kwargs, **kwargs)

    return wrapper

  def copy_kwargs_sunder(func):
    """Pack all of that func's kwargs into __kwargs and pass the kwarguments normally along with _kwargs kwarg which contains all the kwargs in a dict.

    This will error if func receives a literal "__kwargs" kwarg.
    ```py
    >>> a(_kwargs=kwargs, **{"_kwargs": "some_value"})
    ```
    ```txt
    TypeError: __main__.a() got multiple values for keyword argument '_kwargs'
    ```
    """

    @wraps(func)
    def wrapper(*args, **kwargs):
      return func(*args, _kwargs=kwargs, **kwargs)

    return wrapper


if True:  # \/ # @skip_first_call

  def skip_first_call(func):
    func._first_call = True

    def wrapper(*args, **kwargs):
      if func._first_call:
        func._first_call = False
        return None
      else:
        return func(*args, **kwargs)

    return wrapper

  def skip_first_call_async(func):
    func._first_call = True

    async def wrapper(*args, **kwargs):
      if func._first_call:
        func._first_call = False
        return None
      else:
        return await func(*args, **kwargs)

    return wrapper


if True:  # \/ # @with_overrides

  def with_overrides(*overrides: str):
    """For each of the passed in keyword argument names: if one was passed in by the user, use that one, else use `self.{name}`."""

    def decorator(func: Callable):
      def wrapper(self, *args, **kwargs):
        kws = kwargs.copy()
        for o in overrides:
          if o not in kwargs:
            kws[o] = getattr(self, o)

        return func(self, *args, **kws)

      return wrapper

    return decorator
